from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Optional, TypeVar

import astropy.units as u
import h5py
import numpy as np

from opencosmo.dataset import Dataset
from opencosmo.dataset.state import DatasetState
from opencosmo.spatial.healpix import HealPixIndex
from opencosmo.spatial.tree import Tree
from opencosmo.spatial.utils import combine_upwards

if TYPE_CHECKING:
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.spatial import Region

T = TypeVar("T")
GroupedColumnData = dict[str, dict[str, T]]
SpatialIndexData = dict[int, tuple[np.ndarray, int]]


def build_dataset_from_data(
    data: GroupedColumnData[np.ndarray] | h5py.Group,
    header: OpenCosmoHeader,
    region: Region,
    spatial_index_data: SpatialIndexData,
    descriptions: GroupedColumnData[str] = {},
) -> Dataset:
    data_keys = set(data.keys())
    metadata_group = None
    if not data_keys or len(data_keys) > 2 or "data" not in data_keys:
        raise ValueError(
            "Data must have at least one `data` group and at most one metadata group"
        )
    if descriptions and not set(descriptions.keys()).issubset(data.keys()):
        raise ValueError(
            "Descriptions should be organized into the same groups as the data!"
        )

    if isinstance(data, dict):
        data = make_in_memory_h5_file_from_data(data, descriptions)
    if isinstance(spatial_index_data, dict):
        spatial_index_data = make_spatial_index(spatial_index_data)

    tree = Tree(HealPixIndex(), spatial_index_data)
    if len(data_keys) == 2:
        data_keys.remove("data")
        metadata_key = data_keys.pop()
        metadata_group = data[metadata_key]

    new_state = DatasetState.from_group(
        data["data"],
        header,
        header.file.unit_convention,
        region,
        metadata_group=metadata_group,
    )
    return Dataset(header, new_state, tree=tree)


def make_in_memory_h5_file_from_data(
    data: GroupedColumnData[np.ndarray],
    descriptions: Optional[GroupedColumnData[str]] = None,
) -> h5py.File:
    name = uuid.uuid1()
    file = h5py.File(f"{name}.hdf5", "w", driver="core", backing_store=False)
    for group_name, columns in data.items():
        lengths = set(len(c) for c in columns.values())
        if len(lengths) > 1:
            raise ValueError("Columns within a single group must be the same length!")
        group = file.require_group(group_name)
        group_data, group_metadata = split_data_and_metadata(
            columns,
            descriptions.get(group_name, {}) if descriptions is not None else {},
        )

        for column_name, column_data in columns.items():
            group.create_dataset(column_name, data=column_data)
            group[column_name].attrs.update(group_metadata[column_name])
    return file


def make_spatial_index(data: SpatialIndexData):
    """
    allowed input (for now)

    a single level > 0
    """
    if len(data) != 1:
        raise ValueError("Spatial index creation routines should have a single level")
    level = next(iter(data.keys()))
    size, fold_factor = data[level]
    if level <= 0:
        raise ValueError("Data for creating spatial index should include one level > 0")
    name = uuid.uuid1()
    file = h5py.File(f"{name}.hdf5", "w", driver="core", backing_store=False)

    return combine_upwards(size, fold_factor, level, file)


def split_data_and_metadata(data: dict[str, np.ndarray], descriptions: dict[str, str]):
    output_data = {}
    output_metadata = {}
    for colname, coldata in data.items():
        output_data[colname] = coldata
        column_metadata = {}
        if isinstance(coldata, u.Quantity):
            column_metadata["unit"] = str(coldata.unit)
            output_data[colname] = coldata.value
        if colname in descriptions:
            column_metadata["description"] = descriptions[colname]
        output_metadata[colname] = column_metadata
    return output_data, output_metadata



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from .mask import into_array

if TYPE_CHECKING:
    from . import DataIndex


def from_size(size: int):
    return (np.array([0], dtype=np.int64), np.array([size], dtype=np.int64))


def single_chunk(start: int, size: int):
    return (np.array([start], dtype=np.int64), np.array([size], np.int64))


def empty():
    return (np.array([], dtype=np.int64), np.array([], dtype=np.int64))


def from_range(start: int, end: int):
    size = end - start
    return (np.array([start], dtype=np.int64), np.array([size], np.int64))


def concatenate(*indices: DataIndex):
    np.concatenate(list(map(into_array, indices)))



from functools import singledispatch

import numba as nb
import numpy as np
from numpy.typing import NDArray

"""
Implementations for unary operations on indices
"""

SimpleIndex = NDArray[np.int_]
ChunkedIndex = tuple[NDArray[np.int_], NDArray[np.int_]]


def get_length(index: SimpleIndex | ChunkedIndex):
    match index:
        case np.ndarray():
            return len(index)
        case (np.ndarray(), np.ndarray()):
            return int(np.sum(index[1]))
        case _:
            raise TypeError(f"Invalid index type {type(index)}")


def get_range(index: SimpleIndex | ChunkedIndex):
    match index:
        case np.ndarray():
            return __get_simple_range(index)
        case (np.ndarray(), np.ndarray()):
            return __get_chunked_range(*index)
        case _:
            raise ValueError(f"Unknown index type {type(index)}")


@nb.njit
def __get_simple_range(index: SimpleIndex):
    if len(index) == 0:
        return (0, 0)

    min = index[0]
    max = index[0]
    for val in index[1:]:
        if val < min:
            min = val
        if val > max:
            max = val
    return (min, max)


@nb.njit
def __get_chunked_range(starts: NDArray[np.int_], sizes: NDArray[np.int_]):
    if len(starts) == 0:
        return (0, 0)
    min = starts[0]
    max = min + sizes[0]
    for i in range(1, len(starts)):
        start = starts[i]
        end = start + sizes[i]
        if start < min:
            min = start
        if end > max:
            max = end
    return (min, max)



from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb
import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray


def n_in_range(
    index: NDArray[np.int_] | tuple,
    range_starts: int | NDArray[np.int_],
    range_sizes: int | NDArray[np.int_],
):
    range_starts = np.atleast_1d(range_starts)
    range_sizes = np.atleast_1d(range_sizes)
    match index:
        case np.ndarray():
            return __n_in_range_simple(index, range_starts, range_sizes)
        case (np.ndarray(), np.ndarray()):
            return __n_in_range_chunked(*index, range_starts, range_sizes)
        case _:
            raise ValueError(f"Unknown index type {type(index)}")


def __n_in_range_simple(
    index: NDArray[np.int_], start: NDArray[np.int_], size: NDArray[np.int_]
) -> NDArray[np.int_]:
    if len(start) != len(size):
        raise ValueError("Start and size arrays must have the same length")
    if np.any(size < 0):
        raise ValueError("Sizes must greater than or equal to zero")
    if len(index) == 0:
        return np.zeros_like(start)

    ends = start + size
    index_to_search = np.sort(index)
    start_idxs = np.searchsorted(index_to_search, start, "left")
    end_idxs = np.searchsorted(index_to_search, ends, "left")
    return end_idxs - start_idxs


@nb.njit
def __n_in_range_chunked(
    starts: NDArray[np.int_],
    sizes: NDArray[np.int_],
    range_starts: NDArray[np.int_],
    range_sizes: NDArray[np.int_],
) -> NDArray[np.int_]:
    """
    Return the number of elements in this index that fall within
    a specified data range. Used to mask spatial index.


    As with numpy, this is the half-open range [start, end)
    """
    if len(range_starts) != len(range_sizes):
        raise ValueError("Start and size arrays must have the same length")
    if np.any(range_sizes < 0):
        raise ValueError("Sizes must greater than or equal to zero")
    if len(starts) == 0:
        return np.zeros_like(range_starts)

    index_ranges = np.vstack((starts, starts + sizes))
    output = np.zeros_like(range_starts)
    for i in range(len(range_starts)):
        index_chunk_ranges = np.clip(
            index_ranges, a_min=range_starts[i], a_max=range_starts[i] + range_sizes[i]
        )
        output[i] = np.sum(index_chunk_ranges[1] - index_chunk_ranges[0])

    return output



from __future__ import annotations

from functools import reduce
from itertools import chain, product
from typing import TYPE_CHECKING, Iterable, Mapping, Optional

import rustworkx as rx

if TYPE_CHECKING:
    import astropy.units as u
    import numpy as np

    from opencosmo.column.cache import ColumnCache
    from opencosmo.column.column import ConstructedColumn
    from opencosmo.dataset.handler import Hdf5Handler
    from opencosmo.index import DataIndex
    from opencosmo.units.handler import UnitHandler


def build_dependency_graph(
    derived_columns: Mapping[str, ConstructedColumn],
    names_to_keep: Optional[set[str]] = None,
):
    dependency_graph = rx.PyDiGraph()
    all_requires: set[str] = reduce(
        lambda known, dc: known.union(dc.requires), derived_columns.values(), set()
    )
    nodeidx = dependency_graph.add_nodes_from(all_requires)
    nodemap = {name: idx for (name, idx) in zip(all_requires, nodeidx)}

    for target, derived_column in derived_columns.items():
        requires = derived_column.requires
        produces = derived_column.produces
        if produces is None:
            produces = set((target,))
        to_add = list(filter(lambda p: p not in nodemap, produces))
        new_map = dependency_graph.add_nodes_from(to_add)
        nodemap.update({name: idx for (name, idx) in zip(to_add, new_map)})

        requires_idx = (nodemap[r] for r in requires)
        produces_idx = (nodemap[r] for r in produces)

        dependency_graph.add_edges_from_no_data(product(requires_idx, produces_idx))

    if names_to_keep is not None:
        nodes_to_keep = reduce(
            lambda acc, name: acc.union(rx.ancestors(dependency_graph, nodemap[name])),
            names_to_keep,
            {nodemap[name] for name in names_to_keep},
        )
        dependency_graph = dependency_graph.subgraph(list(nodes_to_keep))

    return dependency_graph


def replace_multi_producers(
    graph: rx.PyDiGraph, derived_columns: Mapping[str, ConstructedColumn]
):
    node_map = {name: i for i, name in enumerate(graph.nodes())}
    missing = set(derived_columns.keys()).difference(node_map.keys())
    if not missing:
        return graph
    for missing_column in missing:
        missing_column_produces = derived_columns[missing_column].produces
        assert missing_column_produces is not None
        outputs = [
            node_map[name] for name in missing_column_produces if name in node_map
        ]
        graph.contract_nodes(outputs, missing_column)
    return graph


def validate_derived_columns(
    derived_columns: dict[str, ConstructedColumn],
    known_raw_columns: set[str],
    units: dict[str, u.Unit],
):
    """
    Validate the network of derived columns. This
    """
    dependency_graph = build_dependency_graph(derived_columns)
    if cycle := rx.digraph_find_cycle(dependency_graph):
        names = [dependency_graph[i] for i in cycle]
        raise ValueError(
            f"Found derived columns that depend on each other! Columns: {names}"
        )

    sources = set(
        filter(
            lambda i: not dependency_graph.in_degree(i),
            range(dependency_graph.num_nodes()),
        )
    )
    source_names = map(lambda i: dependency_graph[i], sources)
    if missing := set(source_names).difference(known_raw_columns):
        raise ValueError(f"Tried to derive columns from unknown columns: {missing}")

    dependency_graph = replace_multi_producers(dependency_graph, derived_columns)

    return validate_derived_units(dependency_graph, derived_columns, units)


def validate_derived_units(
    dependency_graph: rx.PyDiGraph,
    derived_columns: dict[str, ConstructedColumn],
    units: dict[str, u.Unit],
):
    output_units: dict[str, Optional[u.Unit]] = {}
    for node in rx.topological_sort(dependency_graph):
        node_name = dependency_graph[node]
        if node_name in units:
            continue
        new_units = derived_columns[node_name].get_units(units)
        if not isinstance(new_units, dict):
            new_units = {node_name: new_units}
        units |= new_units
        output_units |= new_units
    return output_units


def build_derived_columns(
    all_derived_columns: dict[str, ConstructedColumn],
    derived_columns_to_get: set[str],
    cache: ColumnCache,
    hdf5_handler: Hdf5Handler,
    unit_handler: UnitHandler,
    unit_kwargs: dict,
    index: DataIndex,
) -> dict[str, np.ndarray]:
    """
    Build any derived columns that are present in this dataset. Also returns any columns that
    had to be instantiated in order to build these derived columns.
    """
    if not derived_columns_to_get:
        return {}

    column_names: set[str] = reduce(
        lambda known, dc: known.union(dc[1].produces)
        if dc[1].produces is not None
        else known.union((dc[0],)),
        all_derived_columns.items(),
        set(),
    )

    dependency_graph = build_dependency_graph(
        all_derived_columns, derived_columns_to_get
    )
    cached_data = cache.get_columns(dependency_graph.nodes())
    cached_data |= unit_handler.apply_unit_conversions(cached_data, unit_kwargs)

    additional_derived = column_names.difference(cached_data.keys())

    if not additional_derived:
        return cached_data

    columns_to_fetch = (
        set(dependency_graph.nodes())
        .intersection(hdf5_handler.columns)
        .difference(cached_data.keys())
    )

    raw_data = hdf5_handler.get_data(columns_to_fetch)
    data = cached_data | unit_handler.apply_units(raw_data, unit_kwargs)
    dependency_graph = replace_multi_producers(dependency_graph, all_derived_columns)
    new_derived: dict[str, np.ndarray] = {}

    for colidx in rx.topological_sort(dependency_graph):
        colname = dependency_graph[colidx]
        if colname in data:
            continue
        derived_column = all_derived_columns[colname]
        produces = derived_column.produces
        if produces is None:
            produces = set((colname,))
        if all(name in data for name in produces):
            continue
        output = derived_column.evaluate(data, index)
        if isinstance(output, dict):
            data |= output
            new_derived |= output
        else:
            data[colname] = output
            new_derived[colname] = output

    cache.add_data(new_derived)
    return data | new_derived



from __future__ import annotations

from copy import copy
from functools import reduce
from itertools import cycle
from typing import TYPE_CHECKING, Iterable, Optional
from weakref import finalize

import astropy.units as u
import numpy as np
from astropy.table import QTable

from opencosmo.column.cache import ColumnCache
from opencosmo.column.column import DerivedColumn, EvaluatedColumn
from opencosmo.dataset.derived import (
    build_derived_columns,
    validate_derived_columns,
)
from opencosmo.dataset.handler import Hdf5Handler
from opencosmo.dataset.im import resort, validate_in_memory_columns
from opencosmo.index.build import from_size, single_chunk
from opencosmo.index.mask import into_array
from opencosmo.index.unary import get_length, get_range
from opencosmo.io import schemas as ios
from opencosmo.units import UnitConvention
from opencosmo.units.handler import make_unit_handler

if TYPE_CHECKING:
    import h5py
    from astropy import table, units
    from astropy.cosmology import Cosmology
    from numpy.typing import NDArray

    from opencosmo.column.column import ConstructedColumn
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex
    from opencosmo.spatial.protocols import Region
    from opencosmo.units.handler import UnitHandler


def deregister_state(id: int, cache: ColumnCache):
    cache.deregister_column_group(id)


class DatasetState:
    """
    Holds mutable state required by the dataset. Cleans up the dataset to mostly focus
    on very high-level operations. Not a user facing class.
    """

    def __init__(
        self,
        raw_data_handler: Hdf5Handler,
        cache: ColumnCache,
        derived_columns: dict[str, ConstructedColumn],
        unit_handler: UnitHandler,
        header: OpenCosmoHeader,
        columns: set[str],
        region: Region,
        sort_by: Optional[tuple[str, bool]],
    ):
        self.__raw_data_handler = raw_data_handler
        self.__cache = cache
        self.__derived_columns = derived_columns
        self.__unit_handler = unit_handler
        self.__header = header
        self.__columns = columns
        self.__region = region
        self.__sort_by = sort_by
        self.__cache.register_column_group(id(self), self.__columns)
        finalize(self, deregister_state, id(self), self.__cache)

    def __rebuild(self, **updates):
        new = {
            "raw_data_handler": self.__raw_data_handler,
            "cache": self.__cache,
            "derived_columns": self.__derived_columns,
            "unit_handler": self.__unit_handler,
            "header": self.__header,
            "columns": self.__columns,
            "region": self.__region,
            "sort_by": self.__sort_by,
        } | updates
        return DatasetState(**new)

    def __exit__(self, *exec_details):
        return None

    @classmethod
    def from_group(
        cls,
        group: h5py.Group,
        header: OpenCosmoHeader,
        unit_convention: UnitConvention,
        region: Region,
        index: Optional[DataIndex] = None,
        metadata_group: Optional[h5py.Group] = None,
        in_memory: bool = False,
    ):
        handler = Hdf5Handler.from_group(group, index, metadata_group)
        unit_handler = make_unit_handler(handler.data, header, unit_convention)

        columns = set(handler.columns)
        cache = ColumnCache.empty()
        return DatasetState(
            handler,
            cache,
            {},
            unit_handler,
            header,
            columns,
            region,
            None,
        )

    def __len__(self):
        return get_length(self.__raw_data_handler.index)

    @property
    def descriptions(self):
        all_descriptions = (
            {name: col.description for name, col in self.__derived_columns.items()}
            | self.__raw_data_handler.descriptions
            | self.__cache.descriptions
        )
        return {
            name: description
            for name, description in all_descriptions.items()
            if name in self.columns
        }

    @property
    def raw_index(self):
        if (si := self.get_sorted_index()) is not None:
            ni = into_array(self.__raw_data_handler.index)
            return ni[si]

        return self.__raw_data_handler.index

    @property
    def unit_handler(self):
        return self.__unit_handler

    @property
    def convention(self):
        return self.__unit_handler.current_convention

    @property
    def region(self):
        return self.__region

    @property
    def header(self):
        return self.__header

    @property
    def columns(self) -> list[str]:
        return list(self.__columns)

    @property
    def meta_columns(self) -> list[str]:
        return self.__raw_data_handler.metadata_columns

    def get_data(
        self,
        ignore_sort: bool = False,
        metadata_columns: list = [],
        unit_kwargs: dict = {},
    ) -> table.QTable:
        """
        Get the data for a given handler.
        """
        data = self.__build_derived_columns(unit_kwargs)
        cached_data = self.__cache.get_columns(self.columns)
        converted_cached_data = self.__unit_handler.apply_unit_conversions(
            cached_data, unit_kwargs
        )

        data |= cached_data
        if converted_cached_data:
            self.__cache.add_data(converted_cached_data, {}, push_up=False)
            data |= converted_cached_data

        raw_columns = (
            set(self.columns)
            .intersection(self.__raw_data_handler.columns)
            .difference(data.keys())
        )
        if (
            self.__sort_by is not None
            and self.__sort_by[0] in self.__raw_data_handler.columns
        ):
            raw_columns.add(self.__sort_by[0])

        if raw_columns:
            raw_data = self.__raw_data_handler.get_data(raw_columns)
            raw_data = self.__unit_handler.apply_raw_units(raw_data, unit_kwargs)

            if not self.__raw_data_handler.in_memory:
                self.__cache.add_data(raw_data)
            updated_data = self.__unit_handler.apply_unit_conversions(
                raw_data, unit_kwargs
            )
            if updated_data and not self.__raw_data_handler.in_memory:
                self.__cache.add_data(updated_data, push_up=False)
            data |= raw_data | updated_data

        if not set(data.keys()).issuperset(self.columns):
            raise RuntimeError(
                "Some columns are missing from the output! This is likely a bug. Please report it on GitHub"
            )

        # keep ordering

        data_columns = set(data.keys())

        if metadata_columns:
            data.update(self.__raw_data_handler.get_metadata(metadata_columns))

        if not ignore_sort and self.__sort_by is not None:
            sort_by = data[self.__sort_by[0]]
            order = np.argsort(sort_by)
            if self.__sort_by[1]:
                order = order[::-1]

            data = {key: value[order] for key, value in data.items()}

        new_order = [c for c in self.columns]
        if metadata_columns:
            new_order.extend(metadata_columns)

        return {name: data[name] for name in new_order}

    def rows(self, metadata_columns: list = [], unit_kwargs: dict = {}):
        derived_to_collect = (
            set(self.__derived_columns.keys())
            .intersection(self.columns)
            .difference(self.__cache.columns)
        )
        derived_storage: dict[str, list[np.ndarray]] = {
            name: [] for name in derived_to_collect
        }
        total_length = len(self)
        chunk_ranges = [
            (i, min(i + 1000, total_length)) for i in range(0, total_length, 1000)
        ]
        if not chunk_ranges:
            raise StopIteration

        try:
            for start, end in chunk_ranges:
                chunk = self.take_range(start, end)
                data = chunk.get_data(
                    metadata_columns=metadata_columns, unit_kwargs=unit_kwargs
                )
                for name in derived_to_collect:
                    derived_storage[name].append(data[name])

                for i in range(len(chunk)):
                    yield {name: column[i] for name, column in data.items()}
            all_derived = {
                name: np.concatenate(arr) for name, arr in derived_storage.items()
            }
            derived_storage = resort(all_derived, self.get_sorted_index())
            if derived_storage:
                self.__cache.add_data(data)
        except GeneratorExit:
            pass
        except BaseException:
            raise

    def get_metadata(self, columns=[]):
        metadata = self.__raw_data_handler.get_metadata(columns)
        sorted_index = self.get_sorted_index()
        if sorted_index is not None:
            metadata = {name: values[sorted_index] for name, values in metadata.items()}
        return metadata

    def with_mask(self, mask: NDArray[np.bool_]):
        index = np.where(mask)[0]
        new_raw_handler = self.__raw_data_handler.take(index)
        new_cache = self.__cache.take(index)
        return self.__rebuild(
            cache=new_cache,
            raw_data_handler=new_raw_handler,
        )

    def make_schema(self):
        header = self.__header.with_region(self.__region)
        raw_columns = self.__columns.intersection(self.__raw_data_handler.columns)

        schema = self.__raw_data_handler.make_schema(raw_columns, header)
        derived_names = set(self.__derived_columns.keys()).intersection(self.columns)
        derived_data = (
            self.select(derived_names)
            .with_units(self.__unit_handler.base_convention, {}, {}, None, None)
            .get_data(ignore_sort=True)
        )
        column_units = {
            name: self.__unit_handler.base_units[name] for name in raw_columns
        }

        for colname in derived_names:
            coldata = derived_data[colname]
            unit = ""
            if isinstance(coldata, u.Quantity):
                unit = str(coldata.unit)
                coldata = derived_data[colname].value

            attrs = {
                "unit": unit,
                "description": self.__derived_columns[colname].description,
            }
            colschema = ios.ColumnSchema(
                colname, from_size(len(coldata)), coldata, attrs
            )
            schema.add_child(colschema, f"data/{colname}")

        cached_data = self.__cache.get_columns(self.columns)

        for colname, coldata in cached_data.items():
            try:
                data = coldata.value
                unit_str = str(coldata.unit)
            except AttributeError:
                data = coldata
                unit_str = ""
            if colname in schema.columns["data"]:
                continue

            attrs = {}
            attrs["unit"] = unit_str
            attrs["description"] = self.descriptions.get(colname, "None")

            colschema = ios.ColumnSchema(colname, from_size(len(coldata)), data, attrs)
            schema.add_child(colschema, f"data/{colname}")

        return schema

    def with_new_columns(
        self,
        descriptions: dict[str, str] = {},
        **new_columns: DerivedColumn | np.ndarray | units.Quantity,
    ):
        """
        Add a set of derived columns to the dataset. A derived column is a column that
        has been created based on the values in another column.
        """
        from time import time

        derived_update: dict[str, DerivedColumn] = {}
        existing_columns = set(self.columns)

        if inter := existing_columns.intersection(new_columns.keys()):
            raise ValueError(f"Some columns are already in the dataset: {inter}")

        new_derived_columns = {}
        new_in_memory_columns = {}
        new_in_memory_descriptions = {}

        for colname, column in new_columns.items():
            match column:
                case DerivedColumn() | EvaluatedColumn():
                    column.description = descriptions.get(colname, "None")
                    new_derived_columns[colname] = column
                case np.ndarray():
                    if len(column) != len(self):
                        raise ValueError(
                            f"New column {colname} does not have the same length as this dataset!"
                        )
                    new_in_memory_descriptions[colname] = descriptions.get(
                        colname, "None"
                    )
                    new_in_memory_columns[colname] = column
                case _:
                    raise ValueError(
                        f"Got an invalid new column of type {type(column)}"
                    )

        new_unit_handler = self.__unit_handler
        new_cache = self.__cache
        new_derived = copy(self.__derived_columns)
        new_column_names: set[str] = set()

        if new_derived_columns:
            new_units = validate_derived_columns(
                self.__derived_columns | new_derived_columns,
                existing_columns.union(new_in_memory_columns.keys()).difference(
                    self.__derived_columns.keys()
                ),
                self.__unit_handler.base_units,
            )
            new_derived |= new_derived_columns
            for colname, derived in new_derived.items():
                if (prod := derived.produces) is not None:
                    new_column_names |= prod
                else:
                    new_column_names.add(colname)
            new_unit_handler = new_unit_handler.with_new_columns(**new_units)

        if new_in_memory_columns:
            new_unit_handler = validate_in_memory_columns(
                new_in_memory_columns, self.__unit_handler, len(self)
            )
            new_in_memory_columns = resort(
                new_in_memory_columns, self.get_sorted_index()
            )
            new_cache = new_cache.with_data(
                new_in_memory_columns, descriptions=new_in_memory_descriptions
            )
            new_column_names |= set(new_in_memory_columns.keys())

        new_column_names = set(self.columns).union(new_column_names)
        return self.__rebuild(
            cache=new_cache,
            derived_columns=new_derived,
            columns=new_column_names,
            unit_handler=new_unit_handler,
        )

    def __build_derived_columns(self, unit_kwargs: dict) -> table.Table:
        """
        Build any derived columns that are present in this dataset
        """
        if not self.__derived_columns:
            return {}

        all_derived_columns: set[str] = reduce(
            lambda acc, dc: acc.union(
                dc[1].produces if dc[1].produces is not None else {dc[0]}
            ),
            self.__derived_columns.items(),
            set(),
        )
        derived_names = all_derived_columns.intersection(self.columns)
        if self.__sort_by is not None and self.__sort_by[0] in all_derived_columns:
            derived_names.add(self.__sort_by[0])

        dc = build_derived_columns(
            self.__derived_columns,
            derived_names,
            self.__cache,
            self.__raw_data_handler,
            self.__unit_handler,
            unit_kwargs,
            self.__raw_data_handler.index,
        )
        return dc

    def __get_im_columns(self, data: dict, unit_kwargs) -> table.Table:
        im_data = {}
        for colname, column in self.__cache.columns():
            im_data[colname] = column

        return self.__unit_handler.apply_units(im_data, unit_kwargs)

    def with_region(self, region: Region):
        """
        Return the same dataset but with a different region
        """
        return self.__rebuild(region=region)

    def select(self, columns: str | Iterable[str]):
        """
        Select a subset of columns from the dataset. It is possible for a user to select
        a derived column in the dataset, but not the columns it is derived from.
        This class tracks any columns which are required to materialize the dataset but
        are not in the final selection in self.__hidden. When the dataset is
        materialized, the columns in self.__hidden are removed before the data is
        returned to the user.

        """
        if isinstance(columns, str):
            columns = [columns]

        columns = set(columns)
        missing = columns - self.__columns
        if missing:
            raise ValueError(
                f"Tried to select columns that are not in this dataset: {missing}"
            )

        return self.__rebuild(columns=columns)

    def sort_by(self, column_name: str, invert: bool):
        if column_name not in self.columns:
            raise ValueError(f"This dataset has no column {column_name}")

        return self.__rebuild(sort_by=(column_name, invert))

    def get_sorted_index(self):
        if self.__sort_by is not None:
            column = self.select(self.__sort_by[0]).get_data(ignore_sort=True)[
                self.__sort_by[0]
            ]
            sorted = np.argsort(column)
            if self.__sort_by[1]:
                sorted = sorted[::-1]

        else:
            sorted = None

        return sorted

    def take(self, n: int, at: str):
        """
        Take rows from the dataset.
        """

        take_index: DataIndex

        if at == "start":
            return self.take_range(0, n)
        elif at == "end":
            return self.take_range(len(self) - n, len(self))
        elif at == "random":
            row_indices = np.random.choice(len(self), n, replace=False)
            row_indices.sort()

        sorted = self.get_sorted_index()
        if sorted is None:
            take_index = row_indices
        else:
            take_index = np.sort(sorted[row_indices])

        new_handler = self.__raw_data_handler.take(take_index)
        new_cache = self.__cache.take(take_index)

        return self.__rebuild(
            raw_data_handler=new_handler,
            cache=new_cache,
        )

    def take_range(self, start: int, end: int):
        """
        Take a range of rows form the dataset.
        """
        if start < 0 or end < 0:
            raise ValueError("start and end must be positive.")
        if end < start:
            raise ValueError("end must be greater than start.")
        if end > len(self):
            raise ValueError("end must be less than the length of the dataset.")

        sorted = self.get_sorted_index()
        take_index: DataIndex
        if sorted is None:
            take_index = single_chunk(start, end - start)
        else:
            take_index = np.sort(sorted[start:end])

        new_raw_handler = self.__raw_data_handler.take(take_index)
        new_im = self.__cache.take(take_index)
        return self.__rebuild(
            raw_data_handler=new_raw_handler,
            cache=new_im,
        )

    def take_rows(self, rows: DataIndex):
        if len(self) == 0:
            return self
        row_range = get_range(rows)

        if row_range[1] > len(self) or row_range[0] < 0:
            raise ValueError(
                f"Row indices must be between 0 and the length of this dataset!"
            )
        sorted = self.get_sorted_index()
        new_handler = self.__raw_data_handler.take(rows, sorted)
        new_cache = self.__cache.take(rows)

        return self.__rebuild(
            raw_data_handler=new_handler,
            cache=new_cache,
        )

    def with_units(
        self,
        convention: Optional[str],
        conversions: dict[u.Unit, u.Unit],
        columns: dict[str, u.Unit],
        cosmology: Cosmology,
        redshift: float | table.Column,
    ):
        """
        Change the unit convention
        """

        if convention is None:
            convention_ = self.__unit_handler.current_convention
            cache = self.__cache.duplicate()
        else:
            convention_ = UnitConvention(convention)
            cache = self.__cache.without_columns(self.__raw_data_handler.columns)
            cache = cache.without_columns(self.__derived_columns.keys())

        if (
            convention_ == UnitConvention.SCALEFREE
            and UnitConvention(self.header.file.unit_convention)
            != UnitConvention.SCALEFREE
        ):
            raise ValueError(
                f"Cannot convert units with convention {self.header.file.unit_convention} to convention scalefree"
            )
        column_keys = set(columns.keys())
        missing_columns = column_keys - set(self.columns)
        if missing_columns:
            raise ValueError(f"Dataset does not have columns {missing_columns}")

        new_handler = self.__unit_handler.with_convention(convention_).with_conversions(
            conversions, columns
        )
        return self.__rebuild(unit_handler=new_handler, cache=cache)



from __future__ import annotations

from functools import singledispatch
from typing import TYPE_CHECKING

import numba as nb
import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

from opencosmo.index.unary import get_length


def mask(index, boolean_mask):
    match index:
        case np.ndarray():
            return __mask_simple(index, boolean_mask)
        case (np.ndarray(), np.ndarray()):
            return __mask_chunked(index, boolean_mask)
        case _:
            raise TypeError(f"Unknown index type {type(index)}")


def __mask_simple(index: NDArray[np.int_], boolean_mask: NDArray[np.bool_]):
    if (lm := len(boolean_mask)) > len(index):
        raise ValueError(
            "Boolean mask must be less than or equal to the length of the index itself"
        )

    return index[:lm][boolean_mask]


def __mask_chunked(index: tuple, boolean_mask: NDArray[np.bool_]):
    array = into_array(index)
    return array[boolean_mask]


def into_array(index: np.ndarray | tuple):
    if get_length(index) == 0:
        return np.array([], dtype=np.int64)

    match index:
        case np.ndarray():
            return index
        case (np.ndarray(), np.ndarray()):
            return __chunked_into_array(*index)


@nb.njit
def __chunked_into_array(starts: NDArray[np.int_], sizes: NDArray[np.int_]):
    output = np.zeros(np.sum(sizes), dtype=np.int64)
    rs = 0
    for i in range(len(starts)):
        output[rs : rs + sizes[i]] = np.arange(
            starts[i], starts[i] + sizes[i], dtype=np.int64
        )
        rs += sizes[i]
    return output



from __future__ import annotations

from functools import cached_property, reduce
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable, Optional, Self

import astropy.units as u  # type: ignore
import healpy as hp
import healsparse as hsp
import numpy as np
from astropy.table import Column, vstack  # type: ignore

import opencosmo as oc
from opencosmo.column.column import DerivedColumn
from opencosmo.dataset.build import build_dataset_from_data
from opencosmo.evaluate import prepare_kwargs
from opencosmo.index import SimpleIndex, into_array
from opencosmo.io.io import open_single_dataset
from opencosmo.io.schemas import LightconeSchema
from opencosmo.spatial.region import ConeRegion, HealPixRegion

if TYPE_CHECKING:
    from astropy.coordinates import SkyCoord
    from astropy.cosmology import Cosmology
    from astropy.table import Table

    from opencosmo.column.column import ColumnMask
    from opencosmo.dataset import Dataset
    from opencosmo.dataset.build import GroupedColumnData
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.io.io import OpenTarget
    from opencosmo.parameters.hacc import HaccSimulationParameters
    from opencosmo.spatial import Region


def take_from_sorted(
    healpix_map: "HealpixMap", sort_by: str, invert: bool, n: int, at: str | int
):
    column = np.concatenate(
        [ds.select(sort_by).get_data("numpy") for ds in healpix_map.values()]
    )
    if invert:
        column = -column
    sort_index = np.argsort(column)
    if at == "start":
        sort_index = sort_index[:n]
    elif at == "end":
        sort_index = sort_index[-n:]
    elif isinstance(at, int):
        if at + n > len(sort_index) or at < 0:
            raise ValueError(
                "Requested a range that is outside the size of this dataset!"
            )
        sort_index = sort_index[at : at + n]

    sorted_indices = np.sort(sort_index)
    return sorted_indices


class HealpixMap(dict):
    """
    A HealpixMap contains one or more datasets of map format. Each dataset will
    typically contain a different type of data over a specified integrated
    redshift range. The HealpixMap object provides an API identical to the standard
    Dataset API, however the data that is provided is returned in healpix or healsparse
    format, which are different than other opencosmo datasets. This also contains some
    convenience functions for standard operations.
    """

    def __init__(
        self,
        datasets: dict[str, Dataset],
        nside: int,
        nside_lr: int,
        ordering: str,
        full_sky: bool,
        z_range: tuple[float, float],
        hidden: Optional[set[str]] = None,
        ordered_by: Optional[tuple[str, bool]] = None,
        region: Optional[Region] = None,
    ):
        if any("pixel" not in dataset.meta_columns for dataset in datasets.values()):
            raise ValueError("Missing a pixel column for this map!")
        self.update(datasets)
        self.__nside = nside
        self.__nside_lr = nside_lr
        self.__full_sky = full_sky
        self.__z_range = z_range
        self.__ordering = ordering

        columns: set[str] = reduce(
            lambda left, right: left.union(set(right.columns)), self.values(), set()
        )
        if len(columns) != len(next(iter(self.values())).columns):
            raise ValueError("Not all map datasets have the same columns!")

        header = next(iter(self.values())).header
        self.__header = header
        if hidden is None:
            hidden = set()

        self.__hidden = hidden
        self.__ordered_by = ordered_by
        if region is None:
            region = next(iter(self.values())).region
        self.__region = region

    @property
    def nside(self):
        """
        The healpix nside resolution parameter for this map

        Returns
        -------
        dtype: int
        """
        return self.__header.healpix_map["nside"]

    @property
    def pixels(self):
        """
        The healoix pixels that are included in this map
        """
        return next(iter(self.values())).get_metadata(["pixel"])["pixel"]

    @property
    def nside_lr(self):
        """
        The low resolution nside resolution parameter used to
        access this map with healsparse.
        Returns
        -------
        dtype: int
        """
        return self.__header.healpix_map["nside_lr"]

    @property
    def ordering(self):
        """
        The order of pixelization for the map. Either
        NESTED or RING. Maps are currently always saved
        in NESTED format.

        Returns
        -------
        dtype: str
        """
        return self.__header.healpix_map["ordering"]

    @property
    def full_sky(self):
        """
        Whether the map has full-sky coverage or not
        (note if not you must ask for the data in
        healsparse format and not full healpix format)
        Returns
        -------
        dtype: bool
        """
        return self.__header.healpix_map["full_sky"]

    def __repr__(self):
        """
        A basic string representation of the dataset
        """
        length = len(self)

        if len(self) < 10:
            repr_ds = self
            table_head = ""
        else:
            repr_ds = self.take(10, at="start")
            table_head = "First 10 rows:\n"

        table_repr = repr_ds.data.__repr__()
        # remove the first line
        table_repr = table_repr[table_repr.find("\n") + 1 :]
        head = f"OpenCosmo Healpix Map Dataset (length={length}, "
        cosmo_repr = f"Cosmology: {self.cosmology.__repr__()}" + "\n"
        return head + cosmo_repr + table_head + table_repr

    def __len__(self):
        return sum(len(ds) for ds in self.values())

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        for dataset in self.values():
            try:
                dataset.close()
            except ValueError:
                continue

    @property
    def header(self) -> OpenCosmoHeader:
        """
        The header associated with this dataset.

        OpenCosmo headers generally contain information about the original data this
        dataset was produced from, as well as any analysis that was done along
        the way.

        Returns
        -------
        header: opencosmo.header.OpenCosmoHeader

        """
        return self.__header

    @property
    def columns(self) -> list[str]:
        """
        The names of the columns in this map.

        Returns
        -------
        columns: list[str]
        """
        cols = next(iter(self.values())).columns
        cols = list(filter(lambda col: col not in self.__hidden, cols))
        return cols

    @cached_property
    def descriptions(self) -> dict[str, Optional[str]]:
        """
        Return the descriptions (if any) of the columns in this map as a dictonary.
        Columns without a description will be included in the dictionary with a value
        of None

        Returns
        -------

        descriptions : dict[str, str | None]
            The column descriptions
        """
        descriptions = next(iter(self.values())).descriptions
        descriptions = dict(
            filter(lambda kv: kv[0] not in self.__hidden, descriptions.items())
        )
        return descriptions

    @property
    def cosmology(self) -> Cosmology:
        """
        The cosmology of the simulation this dataset is drawn from as
        an astropy.cosmology.Cosmology object.

        Returns
        -------
        cosmology: astropy.cosmology.Cosmology
        """
        return self.__header.cosmology

    @property
    def region(self) -> Region:
        """
        The region this map covers.

        Returns
        -------
        region: opencosmo.spatial.Region

        """
        return self.__region

    @property
    def simulation(self) -> HaccSimulationParameters:
        """
        The parameters of the simulation this dataset is drawn
        from.

        Returns
        -------
        parameters: opencosmo.parameters.hacc.HaccSimulationParameters
        """
        return self.__header.simulation

    @property
    def z_range(self):
        """
        The redshift range of the data which created this map.

        Returns
        -------
        z_range: tuple[float, float]
        """

        return self.__header.healpix_map["z_range"]

    def get_data(self, output="healsparse", nside_out: Optional[int] = None):
        """
        Get the data in this dataset as healsparse map or as healpix maps
        (nest-ordered numpy array). Note that a dataset does not load data from
        disk into memory until this function is called. As a result, you should
        not call this function until you have performed any transformations you
        plan to on the data.

        You can get the data in two formats, "healsparse" (the default) and "healpix".
        "healsparse" format will return the data as a healsparse sparse map.
        "healpix" will return the data as a dictionary of numpy arrays. For map data,
        due to format requirements, no units will be attached to the data itself,
        although these will match the units from the data attributes.


        Parameters
        ----------
        output: str, default="healsparse"
            The format to output the data in

        Returns
        -------
        data: HealsparseMap | Column | dict[str, ndarray] | ndarray
            The data in this dataset.
        """

        if output not in {"healsparse", "healpix"}:
            raise ValueError(f"Unknown output type {output}")

        if nside_out is not None:
            return self.with_resolution(nside_out).get_data(output)

        data = [
            ds.get_data(unpack=False, metadata_columns=["pixel"])
            for ds in self.values()
        ]
        table = vstack(data, join_type="exact")
        table.sort("pixel", reverse=False)

        if output == "healpix":
            if self.__len__() != hp.nside2npix(self.nside):
                raise ValueError(
                    f"healpix type chosen but length of dataset doesn't match nside value"
                )

        if len(table.colnames) == 1:
            table = next(table.itercols())

        if output == "healpix":
            if isinstance(table, (u.Quantity, Column)):
                return table.value
            else:
                table.remove_columns(self.__hidden)
                return {name: col.value for name, col in table.items()}
        elif output == "healsparse":
            dict_maps = {}
            for name, col in table.items():
                if name != "pixel":
                    hsp_out = hsp.HealSparseMap.make_empty(
                        self.nside_lr, self.nside, dtype=np.float32
                    )
                    hsp_out[table["pixel"].value] = (col.value).astype(np.float32)
                    dict_maps[name] = hsp_out
            return dict_maps

    @property
    def data(self):
        """
        Return the data in the dataset in healsparse format. The value of this
        attribute is equivalent to the return value of
        :code:`Dataset.get_data("healsparse")`.


        Returns
        -------
        data : HealsparseMap
            The data in the dataset.

        """
        return self.get_data("healsparse")

    def with_resolution(self, nside):
        """
        Return a copy of the map with a new nside resolution.

        The new resolution must be strictly less than the current
        resolution.

        """
        nside_out = nside
        if nside_out == self.nside:
            return self
        elif nside_out > self.nside:
            raise ValueError(
                "You cannot change the resolution of a map to be higher than its original resolution!"
            )

        data = [
            ds.get_data(unpack=False, metadata_columns=["pixel"])
            for ds in self.values()
        ]
        table = vstack(data, join_type="exact")
        table.sort("pixel", reverse=False)

        if nside_out >= self.__nside:
            raise ValueError(
                f"New nside {nside_out} is greater than or equal to input value: {self.__nside}."
            )
        if not hp.isnsideok(nside_out):
            raise ValueError(f"New nside {nside_out} is invalid.")

        nside_ratio = self.__nside // nside_out
        pixel_lores = table["pixel"].value // (nside_ratio * nside_ratio)

        new_pixels, boundaries = np.unique(pixel_lores, return_index=True)
        counts = np.add.reduceat(np.ones_like(table["pixel"].value), boundaries).astype(
            float
        )

        new_data: GroupedColumnData = {"data": {}, "metadata": {"pixel": new_pixels}}
        for name in self.columns:
            new_data["data"][name] = (
                np.add.reduceat(table[name].value, boundaries) / counts
            ).astype(np.float32)

        new_header = self.header.with_parameters({"map_params/nside": nside_out})

        index_level = 6
        index_nside = 2**index_level
        while index_nside > nside_out:
            index_level -= 1
            index_nside = 2**index_level

        out_npix = hp.nside2npix(nside_out)
        index_npix = hp.nside2npix(index_nside)

        pix_per_idx = out_npix / index_npix
        size = np.full(index_npix, pix_per_idx)

        new_dataset = build_dataset_from_data(
            new_data,
            new_header,
            self.region,
            {index_level: (size, 4)},
            descriptions={
                "data": {
                    key: desc
                    for key, desc in self.descriptions.items()
                    if desc is not None
                }
            },
        )

        return HealpixMap(
            {"data": new_dataset},
            nside_out,
            self.nside_lr,
            self.ordering,
            self.full_sky,
            self.z_range,
            self.__hidden,
            self.__ordered_by,
        )

    @classmethod
    def open(cls, targets: list[OpenTarget], **kwargs):
        datasets: dict[str, Dataset] = {}

        for target in targets:
            ds = open_single_dataset(target)
            # TODO: check if we need some equivalent here
            if not isinstance(ds, HealpixMap) or len(ds.keys()) != 1:
                raise ValueError(
                    "HealpixMap class can only contain datasets (not collections)"
                )
            if target.group.name != "/":
                key = target.group.name.split("/")[-1]
            else:
                key = f"{target.header.healpix_map.z_range}_{target.header.file.data_type}"
            datasets[key] = next(iter(ds.values()))

        return cls(
            datasets,
            ds.nside(),
            ds.nside_lr(),
            ds.ordering(),
            ds.full_sky(),
            ds.z_range(),
        )

    def __map(
        self,
        method,
        *args,
        hidden: Optional[set[str]] = None,
        mapped_arguments: dict[str, dict[str, Any]] = {},
        construct: bool = True,
        **kwargs,
    ):
        """
        This type of collection will only ever be constructed if all the underlying
        datasets have the same data type, so it is always safe to map operations
        across all of them.
        """
        output = {}
        hidden = hidden if hidden is not None else self.__hidden
        for ds_name, dataset in self.items():
            dataset_mapped_arguments = {
                arg_name: args[ds_name] for arg_name, args in mapped_arguments.items()
            }
            output[ds_name] = getattr(dataset, method)(
                *args, **kwargs, **dataset_mapped_arguments
            )

        if construct:
            return HealpixMap(
                output,
                self.nside,
                self.nside_lr,
                self.ordering,
                self.full_sky,
                self.z_range,
                self.__hidden,
                self.__ordered_by,
            )
        return output

    def __map_attribute(self, attribute):
        return {k: getattr(v, attribute) for k, v in self.items()}

    def make_schema(self) -> LightconeSchema:
        schema = LightconeSchema()  # NOTE: so far we're keeping the lightcone schema
        for name, dataset in self.items():
            ds_schema = dataset.make_schema()
            ds_schema.header = ds_schema.header
            schema.add_child(ds_schema, name)
        return schema

    def bound(self, region: Region, inclusive: bool = False):
        """
        Restrict this map to some subregion. Be default this will
        include all pixels whose centers fall within the subregion. You can additionally
        include pixels that overalp without there centers being within the
        specified region by passing :code:`inclusive=True`

        If trying to query in a circular region, consider using
        :py:meth:`cone_search <opencosmo.HealpixMap.cone_search>` for simplicity.

        Parameters
        ----------
        region: opencosmo.spatial.Region
            The region to query.

        incluive: bool, default = Flase
            Whether to include pixels that overlap but whose centers are not in the region1

        Returns
        -------
        new_map: opencosmo.HealpixMap
            The map including the pixels within the region.

        Raises
        ------
        ValueError
            If the query region does not overlap with the coverage of this map
            in
        """
        # The best we can do here is turn
        if not isinstance(region, ConeRegion):
            raise TypeError(
                "Currently only cone regions are supported when performing spatial queries on HealpixMaps"
            )

        vec = hp.ang2vec(region.center.ra.value, region.center.dec.value, lonlat=True)
        pixels = hp.query_disc(
            self.nside,
            vec,
            region.radius.to(u.radian).value,
            inclusive=inclusive,
            nest=self.__ordering,
        )
        new_datasets = {}
        for name, dataset in self.items():
            ds_pixels = dataset.get_metadata(["pixel"])["pixel"]

            rows_to_take = np.where(np.isin(ds_pixels, pixels, assume_unique=True))[0]
            new_datasets[name] = dataset.take_rows(rows_to_take)

        return HealpixMap(
            new_datasets,
            self.nside,
            self.nside_lr,
            self.ordering,
            False,
            self.z_range,
            self.__hidden,
            self.__ordered_by,
            region=HealPixRegion(pixels, self.nside),
        )

    def cone_search(self, center: tuple | SkyCoord, radius: float | u.Quantity):
        """
        Perform a search for objects within some angular distance of some
        given point on the sky. This is a convinience function around
        :py:meth:`bound <opencosmo.HealpixMap.bound>` and is exactly
        equivalent to

        .. code-block:: python

            region = oc.make_cone(center, radius)
            ds = ds.bound(region)

        Parameters
        ----------
        center: tuple | SkyCoord
            The center of the region to search. If a tuple and no units are provided
            assumed to be RA and Dec in degrees.

        radius: float | astropy.units.Quantity
            The angular radius of the region to query. If no units are provided,
            assumed to be degrees.

        Returns
        -------
        new_map: opencosmo.HealpixMap
            The pixels in these maps that fall within the given region.

        """
        region = oc.make_cone(center, radius)
        return self.bound(region)

    def evaluate(
        self,
        func: Callable,
        format: str = "numpy",
        vectorize=False,
        insert=True,
        **evaluate_kwargs,
    ):
        """
        Iterate over the rows in this collection, apply `func` to each, and collect
        the result as new columns in the dataset. You may also choose to simply return thevalues
        instead of inserting them as a column

        This function is the equivalent of :py:meth:`with_new_columns <opencosmo.HealpixMap.with_new_columns>`
        for cases where the new column is not a simple algebraic combination of existing columns. Unlike
        :code:`with_new_columns`, this method will evaluate the results immediately and the resulting
        columns will not change under unit transformations.

        The function should take in arguments with the same name as the columns in this dataset that
        are needed for the computation, and should return a dictionary of output values.
        The dataset will automatically select the needed columns to avoid unnecessarily reading
        data from disk. The new columns will have the same names as the keys of the output dictionary
        See :ref:`Evaluating On Datasets` for more details.

        If vectorize is set to True, the full columns will be pased to the dataset. Otherwise,
        rows will be passed to the function one at a time.

        This function behaves identically to :py:meth:`Dataset.evaluate <opencosmo.Dataset.evaluate>`

        Parameters
        ----------
        func: Callable
            The function to evaluate on the rows in the dataset.

        format: str, default = "numpy"
            The format of the data that is provided to your function. If "astropy", will be a dictionary of
            astropy quantities. If "numpy", will be a dictionary of numpy arrays.

        vectorize: bool, default = False
            Whether to provide the values as full columns (True) or one row at a time (False)

        insert: bool, default = True
            If true, the data will be inserted as a column in this dataset. Otherwise the data will be returned.

        Returns
        -------
        dataset : HealpixMap
            The new lightcone dataset with the evaluated column(s)
        """
        kwargs, iterable_kwargs = prepare_kwargs(len(self), evaluate_kwargs)
        iterable_kwargs_by_dataset = {}
        indices = np.cumsum(np.fromiter((len(ds) for ds in self.values()), dtype=int))[
            :-1
        ]
        for name, arr in iterable_kwargs.items():
            splits = np.array_split(arr, indices)
            iterable_kwargs_by_dataset[name] = dict(zip(self.keys(), splits))

        result = self.__map(
            "evaluate",
            func=func,
            format=format,
            vectorize=vectorize,
            insert=insert,
            mapped_arguments=iterable_kwargs_by_dataset,
            construct=insert,
            **kwargs,
        )
        if next(iter(result.values())) is None:
            return

        if insert:
            assert isinstance(result, HealpixMap)
            return result

        keys = next(iter(result.values())).keys()
        output = {}
        for key in keys:
            output[key] = np.concatenate([r[key] for r in result.values()])
        return output

    def filter(self, *masks: ColumnMask, **kwargs) -> Self:
        """
        Filter the map based on some criteria. See :ref:`Querying Based on Column
        Values` for more information.

        Parameters
        ----------
        *masks : Mask
            The masks to apply to dataset, constructed with :func:`opencosmo.col`

        Returns
        -------
        dataset : Dataset
            The new dataset with the masks applied.

        Raises
        ------
        ValueError
            If the given  refers to columns that are
            not in the dataset, or the  would return zero rows.

        """
        return self.__map("filter", *masks, **kwargs)

    def rows(self) -> Generator[dict[str, float | u.Quantity], None, None]:
        """
        Iterate over the pixels in the map, returning their individual values.
        Rows are returned as a dictionary. For performance, it is recommended
        to first select the columns you need to work with.

        Yields
        -------
        row : dict
            A dictionary of values for each row in the dataset with units.
        """
        yield from chain.from_iterable(v.rows() for v in self.values())

    def select(self, columns: str | Iterable[str]) -> Self:
        """
        Create a new map from a subset of columns in this map.

        Parameters
        ----------
        columns : str or list[str]
            The column or columns to select.

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected columns.

        Raises
        ------
        ValueError
            If any of the given columns are not in the dataset.
        """

        if isinstance(columns, str):
            columns = [columns]
        columns = set(columns)
        hidden = self.__hidden

        if self.__ordered_by is not None and self.__ordered_by[0] not in columns:
            columns.add(self.__ordered_by[0])

        return self.__map("select", columns, hidden=hidden)

    def drop(self, columns: str | Iterable[str]) -> Self:
        """
        Produce a new dataset by dropping columns from this map.

        Parameters
        ----------
        columns : str or list[str]
            The column or columns to drop.

        Returns
        -------
        dataset : Dataset
            The new dataset without the dropped columns

        Raises
        ------
        ValueError
            If any of the given columns are not in the dataset.
        """
        if isinstance(columns, str):
            columns = [columns]

        dropped_columns = set(columns)
        current_columns = set(self.columns)
        if missing := dropped_columns.difference(current_columns):
            raise ValueError(
                f"Tried to drop columns that are not in this dataset: {missing}"
            )
        kept_columns = current_columns - dropped_columns
        return self.select(kept_columns)

    def take(self, n: int, at: str = "random") -> "HealpixMap":
        """
        Create a new dataset from some number of rows from this map.

        Can take the first n rows, the last n rows, or n random rows
        depending on the value of 'at'.

        Parameters
        ----------
        n : int
            The number of rows to take.
        at : str
            Where to take the rows from. One of "start", "end", or "random".
            The default is "random".

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected rows.

        Raises
        ------
        ValueError
            If n is negative or greater than the number of rows in the dataset,
            or if 'at' is invalid.

        """
        if n > len(self):
            raise ValueError(
                "Number of rows to take must be less than number of rows in dataset"
            )
        if at == "random":
            rs = 0
            indices = np.random.choice(len(self), n, replace=False)
            indices = np.sort(indices)
            return self.__take_rows(indices)

        elif self.__ordered_by is not None:
            index = take_from_sorted(self, *self.__ordered_by, n=n, at=at)
            return self.__take_rows(index)
        elif at == "start":
            return self.take_range(0, n)
        elif at == "end":
            return self.take_range(len(self) - n, len(self))
        else:
            raise ValueError(
                f'"at" should be one of ("start", "end", "random", got {at}'
            )

    def take_range(self, start: int, end: int):
        if start < 0 or end > len(self):
            raise ValueError("Got row indices that are out of range!")

        if self.__ordered_by is not None:
            indices = take_from_sorted(self, *self.__ordered_by, end - start, at=start)
            return self.__take_rows(indices)

        ends = np.cumsum(np.fromiter((len(ds) for ds in self.values()), dtype=int))
        starts = np.insert(ends, 0, 0)[:-1]
        clipped_starts = np.clip(starts, a_min=start, a_max=None)
        clipped_ends = np.clip(ends, a_min=None, a_max=end)

        output = {}
        for i, (name, dataset) in enumerate(self.items()):
            if starts[i] == clipped_starts[i] and ends[i] == clipped_ends[i]:
                output[name] = dataset
            elif clipped_starts[i] >= clipped_ends[i]:
                continue
            else:
                output[name] = dataset.take_range(
                    clipped_starts[i] - starts[i], clipped_ends[i] - starts[i]
                )
        return HealpixMap(
            output,
            self.nside,
            self.nside_lr,
            self.ordering,
            self.full_sky,
            self.z_range,
            self.__hidden,
            self.__ordered_by,
        )

    def take_rows(self, rows: np.ndarray):
        """
        Take the rows of a map specified by the :code:`rows` argument.
        :code:`rows` should be an array of integers. Note that for healpix
        maps the rows refers to the pixel indices.

        Parameters
        ----------
        rows : np.ndarray[int]
            The indices of the rows to take.

        Returns
        -------
        dataset: The dataset with only the specified rows included

        Raises:
        -------
        ValueError:
            If any of the indices is less than 0 or greater than the length of the
            map.

        """
        rows = np.sort(rows)
        if rows[-1] >= len(self) or rows[0] < 0:
            raise ValueError(
                "Rows must be between 0 and the length of this dataset - 1"
            )
        if self.__ordered_by is not None:
            data = np.concatenate(
                [
                    ds.select(self.__ordered_by[0]).get_data("numpy")
                    for ds in self.values()
                ]
            )
            if self.__ordered_by[1]:
                data = -data
            sort_index = np.argsort(data)
            rows = sort_index[rows]
            rows.sort()

        return self.__take_rows(rows)

    def __take_rows(self, rows: np.ndarray):
        """
        Takes rows from this map while ignoring sort. "rows" is assumed to be sorted.
        For internal use only.
        """
        ds_ends = np.cumsum(np.fromiter((len(ds) for ds in self.values()), dtype=int))
        partitions = np.searchsorted(rows, ds_ends)
        splits = np.split(rows, partitions)
        output = {**self}
        rs = 0
        for split, (name, dataset) in zip(splits, self.items()):
            if len(split) > 0:
                output[name] = dataset.take_rows(split - rs)
            rs += len(dataset)
        return HealpixMap(
            output,
            self.nside,
            self.nside_lr,
            self.ordering,
            self.full_sky,
            self.z_range,
            self.__hidden,
            self.__ordered_by,
        )

    def with_new_columns(
        self,
        descriptions: str | dict[str, str] = {},
        **columns: DerivedColumn | np.ndarray | u.Quantity,
    ):
        """
        Create a new map with additional columns. These new columns can be derived
        from columns already in the dataset, or a numpy array.  See :ref:`Adding Custom Columns`
        and :py:meth:`Dataset.with_new_columns <opencosmo.Dataset.with_new_columns>`
        for examples.

        Parameters
        ----------
        descriptions : str | dict[str, str], optional
            A description for the new columns. These descriptions will be accessible through
            :py:attr:`HealpixMap.descriptions <opencosmo.HealpixMap.descriptions>`. If a dictionary,
            should have keys matching the column names.

        ** columns : opencosmo.DerivedColumn | np.ndarray | u.quantity
            The new columns

        Returns
        -------
        dataset : opencosmo.Dataset
            This dataset with the columns added

        """

        derived = {}
        raw = {}
        for name, column in columns.items():
            if isinstance(column, DerivedColumn):
                derived[name] = column
            elif len(column) != len(self):
                raise ValueError(
                    f"New column {name} has length {len(column)} but this dataset "
                    f"has length {len(self)}"
                )
            else:
                raw[name] = column

        split_points = np.cumsum([len(ds) for ds in self.values()])
        split_points = np.insert(0, 0, split_points)[:-1]
        raw_split = {name: np.split(arr, split_points) for name, arr in raw.items()}
        new_datasets = {}
        for i, (ds_name, ds) in enumerate(self.items()):
            raw_columns = {name: arrs[i] for name, arrs in raw_split.items()}
            columns_input = raw_columns | derived
            new_dataset = ds.with_new_columns(descriptions, **columns_input)
            new_datasets[ds_name] = new_dataset
        return HealpixMap(
            new_datasets,
            self.nside,
            self.nside_lr,
            self.ordering,
            self.full_sky,
            self.z_range,
            self.__hidden,
            self.__ordered_by,
        )

    def sort_by(self, column: str, invert: bool = False):
        """
        Sort this map by the values in a given column. By default sorting is in
        ascending order (least to greatest). Pass invert = True to sort in descending
        order (greatest to least).

        This is not generally particular useful in map queries, but can be used to
        enforce ordering schemes or find outlier pixels.

        Parameters
        ----------
        column : str
            The column in the map dataset to
            order the collection by.

        invert : bool, default = False
            If False (the default), ordering will be from least to greatest.
            Otherwise greatest to least.

        Returns
        -------
        result : Dataset
            A new Dataset ordered by the given column.


        """

        if column not in self.columns:
            raise ValueError(f"Column {column} does not exist in this dataset!")
        return HealpixMap(
            dict(self),
            self.nside,
            self.nside_lr,
            self.ordering,
            self.full_sky,
            self.z_range,
            self.__hidden,
            self.__ordered_by,
        )

    def with_units(
        self,
        convention: Optional[str] = None,
        conversions: dict[u.Unit, u.Unit] = {},
        **columns: u.Unit,
    ) -> Self:
        r"""
        Unit conversion is usually supported for OpenCosmo datasets, however maps tend to be integrated
        quantities over a range of redshifts which correspond to observed units so applying unit conversions
        is not generally easy or appropriate.
        """

        raise NotImplementedError(
            "Unit conversions not supported on maps, these are integrated over redshift so conversions are non-trivial!"
        )
